# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import re
from knack.util import CLIError
from knack.log import get_logger
from .custom import get_docker_command
from ._docker_utils import _get_aad_token
from .helm import get_helm_command
from ._utils import get_registry_by_name, resolve_identity_client_id
from ._errors import ErrorClass

logger = get_logger(__name__)

DOCKER_PULL_SUCCEEDED = "Downloaded newer image for {}"
DOCKER_IMAGE_UP_TO_DATE = "Image is up to date for {}"
IMAGE = "mcr.microsoft.com/mcr/hello-world:latest"
FAQ_MESSAGE = "\nPlease refer to https://aka.ms/acr/health-check for more information."
ERROR_MSG_DEEP_LINK = "\nPlease refer to https://aka.ms/acr/errors#{} for more information."
MIN_HELM_VERSION = "2.11.0"
HELM_VERSION_REGEX = re.compile(r'(SemVer|Version):"v([.\d]+)"')
ACR_CHECK_HEALTH_MSG = "Try running 'az acr check-health -n {} --yes' to diagnose this issue."
RECOMMENDED_NOTARY_VERSION = "0.6.0"
NOTARY_VERSION_REGEX = re.compile(r'Version:\s+([.\d]+)')
DOCKER_PULL_WRONG_PLATFORM = 'cannot be used on this platform'


# Utilities functions
def print_pass(message):
    logger.warning("%s : OK", str(message))


def _handle_error(error, ignore_errors):
    if ignore_errors:
        logger.error(error.get_error_message())
    else:
        error_msg = ERROR_MSG_DEEP_LINK.format(error.error_title.lower())
        raise CLIError(error.get_error_message(error_msg))


def _subprocess_communicate(command_parts, shell=False):
    from subprocess import PIPE, Popen, CalledProcessError
    output, stderr = "", ""

    try:
        p = Popen(command_parts, stdout=PIPE, stderr=PIPE, shell=shell)
        output, stderr = p.communicate()
        output = output.decode('UTF-8').rstrip()
        stderr = stderr.decode('UTF-8').rstrip()
    except CalledProcessError as e:
        stderr = str(e)

    warning = None
    if stderr.lower().startswith("warning"):
        warning = stderr
        stderr = None

    if stderr:
        stderr = "Failed to run command '{}'. {}".format(
            ' '.join(command_parts),
            stderr
        )
    return output, warning, stderr


# Checks for the environment
# Checks docker command, docker daemon, docker version and docker pull
def _get_docker_status_and_version(ignore_errors, yes):
    from ._errors import DOCKER_DAEMON_ERROR, DOCKER_PULL_ERROR, DOCKER_VERSION_ERROR

    # Docker command and docker daemon check
    docker_command, error = get_docker_command(is_diagnostics_context=True)
    docker_daemon_available = True

    if error:
        _handle_error(error, ignore_errors)
        if error.error_title != DOCKER_DAEMON_ERROR.error_title:
            return  # We cannot proceed if the error is unexpected or with docker command
        docker_daemon_available = False

    if docker_daemon_available:
        logger.warning("Docker daemon status: available")

    # Docker version check
    output, warning, stderr = _subprocess_communicate(
        [docker_command, "version", "--format", "'Docker version {{.Server.Version}}, "
         "build {{.Server.GitCommit}}, platform {{.Server.Os}}/{{.Server.Arch}}'"])
    if stderr:
        _handle_error(DOCKER_VERSION_ERROR.append_error_message(stderr), ignore_errors)
    else:
        if warning:
            logger.warning(warning)
        logger.warning("Docker version: %s", output)

    # Docker pull check - only if docker daemon is available
    if docker_daemon_available:
        if not yes:
            from knack.prompting import prompt_y_n
            confirmation = prompt_y_n("This will pull the image {}. Proceed?".format(IMAGE))
            if not confirmation:
                logger.warning("Skipping pull check.")
                return

        output, warning, stderr = _subprocess_communicate([docker_command, "pull", IMAGE])

        if stderr:
            if DOCKER_PULL_WRONG_PLATFORM in stderr:
                print_pass("Docker pull of '{}'".format(IMAGE))
                logger.warning("Image '%s' can be pulled but cannot be used on this platform", IMAGE)
            else:
                _handle_error(DOCKER_PULL_ERROR.append_error_message(stderr), ignore_errors)
        else:
            if warning:
                logger.warning(warning)
            if output.find(DOCKER_PULL_SUCCEEDED.format(IMAGE)) != -1 or \
               output.find(DOCKER_IMAGE_UP_TO_DATE.format(IMAGE)) != -1:
                print_pass("Docker pull of '{}'".format(IMAGE))
            else:
                _handle_error(DOCKER_PULL_ERROR, ignore_errors)


# Get current CLI version
def _get_cli_version():
    from pkg_resources import working_set

    # working_set.by_key is a dictionary with component names as key
    cli_component_name = "azure-cli"
    cli_version = "not found"

    if cli_component_name in working_set.by_key:
        cli_version = working_set.by_key[cli_component_name].version

    logger.warning('Azure CLI version: %s', cli_version)


# Get helm versions
def _get_helm_version(ignore_errors):
    from ._errors import HELM_VERSION_ERROR
    from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module

    # Helm command check
    helm_command, error = get_helm_command(is_diagnostics_context=True)

    if error:
        _handle_error(error, ignore_errors)
        return

    # Helm version check
    output, warning, stderr = _subprocess_communicate([helm_command, "version", "--client"])

    if stderr:
        _handle_error(HELM_VERSION_ERROR.append_error_message(stderr), ignore_errors)
        return

    if warning:
        logger.warning(warning)

    # Retrieve the helm version if regex pattern is found
    match_obj = HELM_VERSION_REGEX.search(output)
    if match_obj:
        output = match_obj.group(2)

    logger.warning("Helm version: %s", output)

    # Display an error message if the current helm version < min required version
    if match_obj and LooseVersion(output) < LooseVersion(MIN_HELM_VERSION):
        obsolete_ver_error = HELM_VERSION_ERROR.set_error_message(
            "Current Helm client version is not recommended. Please upgrade your Helm client to at least version {}."
            .format(MIN_HELM_VERSION))
        _handle_error(obsolete_ver_error, ignore_errors)


def _get_notary_version(ignore_errors):
    from ._errors import NOTARY_VERSION_ERROR
    from .notary import get_notary_command
    from distutils.version import LooseVersion  # pylint: disable=import-error,no-name-in-module

    # Notary command check
    notary_command, error = get_notary_command(is_diagnostics_context=True)

    if error:
        _handle_error(error, ignore_errors)
        return

    # Notary version check
    output, warning, stderr = _subprocess_communicate([notary_command, "version"])

    if stderr:
        _handle_error(NOTARY_VERSION_ERROR.append_error_message(stderr), ignore_errors)
        return

    if warning:
        logger.warning(warning)

    # Retrieve the notary version if regex pattern is found
    match_obj = NOTARY_VERSION_REGEX.search(output)
    if match_obj:
        output = match_obj.group(1)

    logger.warning("Notary version: %s", output)

    # Display error if the current version does not match the recommended version
    if match_obj and LooseVersion(output) != LooseVersion(RECOMMENDED_NOTARY_VERSION):
        version_msg = "upgrade"
        if LooseVersion(output) > LooseVersion(RECOMMENDED_NOTARY_VERSION):
            version_msg = "downgrade"
        obsolete_ver_error = NOTARY_VERSION_ERROR.set_error_message(
            "Current notary version is not recommended. Please {} your notary client to version {}."
            .format(version_msg, RECOMMENDED_NOTARY_VERSION))
        _handle_error(obsolete_ver_error, ignore_errors)


# Checks for the connectivity
# Check DNS lookup and access to challenge endpoint
def _get_registry_status(login_server, registry_name, ignore_errors):
    import socket

    registry_ip = None

    try:
        registry_ip = socket.gethostbyname(login_server)
    except (socket.gaierror, UnicodeError):
        # capture UnicodeError for https://github.com/Azure/azure-cli/issues/12936
        pass

    if not registry_ip:
        from ._errors import CONNECTIVITY_DNS_ERROR
        _handle_error(CONNECTIVITY_DNS_ERROR.format_error_message(login_server), ignore_errors)
        return False

    print_pass("DNS lookup to {} at IP {}".format(login_server, registry_ip))

    import requests
    from requests.exceptions import SSLError, RequestException
    from azure.cli.core.util import should_disable_connection_verify

    try:
        challenge = requests.get('https://' + login_server + '/v2/', verify=(not should_disable_connection_verify()))
    except SSLError:
        from ._errors import CONNECTIVITY_SSL_ERROR
        _handle_error(CONNECTIVITY_SSL_ERROR.format_error_message(login_server), ignore_errors)
        return False
    except RequestException:
        from ._errors import CONNECTIVITY_CHALLENGE_ERROR
        _handle_error(CONNECTIVITY_CHALLENGE_ERROR.format_error_message(login_server), ignore_errors)
        return False

    if challenge.status_code == 403:
        from ._errors import CONNECTIVITY_FORBIDDEN_ERROR
        _handle_error(CONNECTIVITY_FORBIDDEN_ERROR.format_error_message(login_server, registry_name), ignore_errors)
        return False
    return True


def _get_endpoint_and_token_status(cmd, login_server, ignore_errors):
    from ._errors import CONNECTIVITY_CHALLENGE_ERROR, CONNECTIVITY_AAD_LOGIN_ERROR, \
        CONNECTIVITY_REFRESH_TOKEN_ERROR, CONNECTIVITY_ACCESS_TOKEN_ERROR

    # Check access to login endpoint
    url = 'https://' + login_server + '/v2/'
    result_from_token = _get_aad_token(cmd.cli_ctx, login_server, False, is_diagnostics_context=True)

    if isinstance(result_from_token, ErrorClass):
        if result_from_token.error_title == CONNECTIVITY_CHALLENGE_ERROR.error_title:
            _handle_error(result_from_token, ignore_errors)
            return

        print_pass("Challenge endpoint {}".format(url))

        if result_from_token.error_title == CONNECTIVITY_AAD_LOGIN_ERROR.error_title:
            _handle_error(result_from_token, ignore_errors)
            return

        if result_from_token.error_title == CONNECTIVITY_REFRESH_TOKEN_ERROR.error_title:
            _handle_error(result_from_token, ignore_errors)
            return

        print_pass("Fetch refresh token for registry '{}'".format(login_server))

        if result_from_token.error_title == CONNECTIVITY_ACCESS_TOKEN_ERROR.error_title:
            _handle_error(result_from_token, ignore_errors)
            return

        print_pass("Fetch access token for registry '{}'".format(login_server))

        return

    # If return is not of type ErrorClass, then it is the token
    print_pass("Challenge endpoint {}".format(url))
    print_pass("Fetch refresh token for registry '{}'".format(login_server))
    print_pass("Fetch access token for registry '{}'".format(login_server))


def _check_registry_health(cmd, registry_name, ignore_errors):
    if registry_name is None:
        logger.warning("Registry name must be provided to check connectivity.")
        return

    registry = None
    # Connectivity
    try:
        registry, _ = get_registry_by_name(cmd.cli_ctx, registry_name)
        login_server = registry.login_server.rstrip('/')
    except CLIError:
        from ._docker_utils import get_login_server_suffix
        suffix = get_login_server_suffix(cmd.cli_ctx)

        if not suffix:
            from ._errors import LOGIN_SERVER_ERROR
            _handle_error(LOGIN_SERVER_ERROR.format_error_message(registry_name), ignore_errors)
            return

        login_server = registry_name + suffix

    status_validated = _get_registry_status(login_server, registry_name, ignore_errors)
    if status_validated:
        _get_endpoint_and_token_status(cmd, login_server, ignore_errors)

    # CMK settings
    if registry and registry.encryption and registry.encryption.key_vault_properties:  # pylint: disable=too-many-nested-blocks
        client_id = registry.encryption.key_vault_properties.identity
        valid_identity = False
        if registry.identity:
            valid_identity = (client_id == 'system') and bool(registry.identity.principal_id)  # use system identity?
            if not valid_identity and registry.identity.user_assigned_identities:
                for k, v in registry.identity.user_assigned_identities.items():
                    if v.client_id == client_id:
                        from msrestazure.azure_exceptions import CloudError
                        try:
                            valid_identity = (resolve_identity_client_id(cmd.cli_ctx, k) == client_id)
                        except CloudError:
                            pass
        if not valid_identity:
            from ._errors import CMK_MANAGED_IDENTITY_ERROR
            _handle_error(CMK_MANAGED_IDENTITY_ERROR.format_error_message(registry_name), ignore_errors)


# General command
def acr_check_health(cmd,  # pylint: disable useless-return
                     ignore_errors=False,
                     yes=False,
                     registry_name=None):
    from azure.cli.core.util import in_cloud_console
    in_cloud_console = in_cloud_console()
    if in_cloud_console:
        logger.warning("Environment checks are not supported in Azure Cloud Shell.")
    else:
        _get_docker_status_and_version(ignore_errors, yes)
        _get_cli_version()

    _check_registry_health(cmd, registry_name, ignore_errors)

    if not in_cloud_console:
        _get_helm_version(ignore_errors)
        _get_notary_version(ignore_errors)

    logger.warning(FAQ_MESSAGE)



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: disable=too-many-lines


def sshkey_create(client,
                  resource_group_name,
                  ssh_public_key_name,
                  location,
                  tags=None,
                  public_key=None):
    parameters = {
        'location': location,
        'tags': tags,
        'public_key': public_key
    }
    client.create(resource_group_name=resource_group_name,
                  ssh_public_key_name=ssh_public_key_name,
                  parameters=parameters)
    if public_key is None:  # Generate one if public key is None
        client.generate_key_pair(resource_group_name=resource_group_name,
                                 ssh_public_key_name=ssh_public_key_name)
    return client.get(resource_group_name=resource_group_name,
                      ssh_public_key_name=ssh_public_key_name)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.decorators import Completer
from azure.cli.core.commands.client_factory import get_subscription_id

from ._client_factory import cf_policy_insights


@Completer
def get_policy_remediation_completion_list(cmd, prefix, namespace, **kwargs):  # pylint: disable=unused-argument
    client = cf_policy_insights(cmd.cli_ctx)
    sub = get_subscription_id(cmd.cli_ctx)
    rg = getattr(namespace, 'resource_group_name', None)
    management_group = getattr(namespace, 'management_group_name', None)
    if rg:
        result = client.remediations.list_for_resource_group(subscription_id=sub, resource_group_name=rg)
    elif management_group:
        result = client.remediations.list_for_management_group(management_group_id=management_group)
    else:
        result = client.remediations.list_for_subscription(subscription_id=sub)

    return [i.name for i in result]


@Completer
def get_policy_metadata_completion_list(cmd, prefix, namespace, **kwargs):  # pylint: disable=unused-argument
    client = cf_policy_insights(cmd.cli_ctx).policy_metadata

    from azure.mgmt.policyinsights.models import QueryOptions
    query_options = QueryOptions(top=2000)

    return [metadata.name for metadata in client.list(query_options) if metadata.name.startswith(prefix)]



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.util import CLIError


def validate_include_or_exclude(namespace):
    if namespace.include and namespace.exclude:
        raise CLIError('usage error: --include | --exclude')


def validate_ids_or_names(namespace):
    if namespace.ids and (namespace.recommendation_name or namespace.resource_group_name):
        raise CLIError('usage error: --ids | --name [--resource-group]')


def validate_ids_or_resource_group(namespace):
    if namespace.ids and namespace.resource_group_name:
        raise CLIError('usage error: --ids | --resource-group')


def validate_threshold_or_resource_group(namespace):
    if namespace.low_cpu_threshold and namespace.resource_group_name:
        raise CLIError('usage error: --low-cpu-threshold | --resource-group')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import ipaddress
import json
import re
import uuid

from azure.cli.core.commands.client_factory import get_mgmt_service_client
from azure.cli.core.commands.client_factory import get_subscription_id
from azure.cli.core.profiles import ResourceType
from azure.cli.core.azclierror import CLIInternalError, InvalidArgumentValueError, \
    RequiredArgumentMissingError
from knack.log import get_logger
from msrestazure.tools import is_valid_resource_id
from msrestazure.tools import parse_resource_id
from msrestazure.tools import resource_id

logger = get_logger(__name__)


def validate_cidr(key):
    def _validate_cidr(namespace):
        cidr = getattr(namespace, key)
        if cidr is not None:
            try:
                ipaddress.IPv4Network(cidr)
            except ValueError:
                raise InvalidArgumentValueError("Invalid --%s '%s'." %
                                                (key.replace('_', '-'), cidr))

    return _validate_cidr


def validate_client_id(namespace):
    if namespace.client_id is not None:
        try:
            uuid.UUID(namespace.client_id)
        except ValueError:
            raise InvalidArgumentValueError("Invalid --client-id '%s'." % namespace.client_id)

        if namespace.client_secret is None or not str(namespace.client_secret):
            raise RequiredArgumentMissingError('Must specify --client-secret with --client-id.')


def validate_client_secret(namespace):
    if namespace.client_secret is not None:
        if namespace.client_id is None or not str(namespace.client_id):
            raise RequiredArgumentMissingError('Must specify --client-id with --client-secret.')


def validate_cluster_resource_group(cmd, namespace):
    if namespace.cluster_resource_group is not None:
        client = get_mgmt_service_client(
            cmd.cli_ctx, ResourceType.MGMT_RESOURCE_RESOURCES)

        if client.resource_groups.check_existence(namespace.cluster_resource_group):
            raise InvalidArgumentValueError(
                "Invalid --cluster-resource-group '%s': resource group must not exist." %
                namespace.cluster_resource_group)


def validate_domain(namespace):
    if namespace.domain is not None:
        if not re.match(r'^' +
                        r'([a-z0-9]|[a-z0-9][-a-z0-9]{0,61}[a-z0-9])' +
                        r'(\.([a-z0-9]|[a-z0-9][-a-z0-9]{0,61}[a-z0-9]))*' +
                        r'$', namespace.domain):
            raise InvalidArgumentValueError("Invalid --domain '%s'." %
                                            namespace.domain)


def validate_pull_secret(namespace):
    if namespace.pull_secret is None:
        # TODO: add aka.ms link here
        warning = "No --pull-secret provided: cluster will not include samples or operators from " + \
            "Red Hat or from certified partners."

        logger.warning(warning)

    else:
        try:
            if not isinstance(json.loads(namespace.pull_secret), dict):
                raise Exception()
        except:
            raise InvalidArgumentValueError("Invalid --pull-secret.")


def validate_subnet(key):
    def _validate_subnet(cmd, namespace):
        from azure.core.exceptions import HttpResponseError
        subnet = getattr(namespace, key)

        if not is_valid_resource_id(subnet):
            if not namespace.vnet:
                raise RequiredArgumentMissingError('Must specify --vnet if --%s is not an id.' %
                                                   key.replace('_', '-'))

            validate_vnet(cmd, namespace)

            subnet = namespace.vnet + '/subnets/' + subnet
            setattr(namespace, key, subnet)

        parts = parse_resource_id(subnet)

        if parts['subscription'] != get_subscription_id(cmd.cli_ctx):
            raise InvalidArgumentValueError("--%s subscription '%s' must equal cluster subscription." %
                                            (key.replace('_', '-'), parts["subscription"]))

        if parts['namespace'].lower() != 'microsoft.network':
            raise InvalidArgumentValueError("--%s namespace '%s' must equal Microsoft.Network." %
                                            (key.replace('_', '-'), parts["namespace"]))

        if parts['type'].lower() != 'virtualnetworks':
            raise InvalidArgumentValueError("--%s type '%s' must equal virtualNetworks." %
                                            (key.replace('_', '-'), parts["type"]))

        if parts['last_child_num'] != 1:
            raise InvalidArgumentValueError("--%s '%s' must have one child." %
                                            (key.replace('_', '-'), subnet))

        if 'child_namespace_1' in parts:
            raise InvalidArgumentValueError("--%s '%s' must not have child namespace." %
                                            (key.replace('_', '-'), subnet))

        if parts['child_type_1'].lower() != 'subnets':
            raise InvalidArgumentValueError("--%s child type '%s' must equal subnets." %
                                            (key.replace('_', '-'), subnet))

        client = get_mgmt_service_client(
            cmd.cli_ctx, ResourceType.MGMT_NETWORK)
        try:
            client.subnets.get(parts['resource_group'],
                               parts['name'], parts['child_name_1'])
        except HttpResponseError as err:
            raise CLIInternalError(err.message)

    return _validate_subnet


def validate_subnets(master_subnet, worker_subnet):
    master_parts = parse_resource_id(master_subnet)
    worker_parts = parse_resource_id(worker_subnet)

    if master_parts['resource_group'].lower() != worker_parts['resource_group'].lower():
        raise InvalidArgumentValueError("--master-subnet resource group '%s' must equal "
                                        "--worker-subnet resource group '%s'." %
                                        (master_parts['resource_group'], worker_parts['resource_group']))

    if master_parts['name'].lower() != worker_parts['name'].lower():
        raise InvalidArgumentValueError("--master-subnet vnet name '%s' must equal --worker-subnet vnet name '%s'." %
                                        (master_parts['name'], worker_parts['name']))

    if master_parts['child_name_1'].lower() == worker_parts['child_name_1'].lower():
        raise InvalidArgumentValueError("--master-subnet name '%s' must not equal --worker-subnet name '%s'." %
                                        (master_parts['child_name_1'], worker_parts['child_name_1']))

    return resource_id(
        subscription=master_parts['subscription'],
        resource_group=master_parts['resource_group'],
        namespace='Microsoft.Network',
        type='virtualNetworks',
        name=master_parts['name'],
    )


def validate_visibility(key):
    def _validate_visibility(namespace):
        visibility = getattr(namespace, key)
        if visibility is not None:
            visibility = visibility.capitalize()
            if visibility not in ['Private', 'Public']:
                raise InvalidArgumentValueError("Invalid --%s '%s'." %
                                                (key.replace('_', '-'), visibility))

    return _validate_visibility


def validate_vnet(cmd, namespace):
    validate_vnet_resource_group_name(namespace)

    if namespace.vnet:
        if not is_valid_resource_id(namespace.vnet):
            namespace.vnet = resource_id(
                subscription=get_subscription_id(cmd.cli_ctx),
                resource_group=namespace.vnet_resource_group_name,
                namespace='Microsoft.Network',
                type='virtualNetworks',
                name=namespace.vnet,
            )


def validate_vnet_resource_group_name(namespace):
    if not namespace.vnet_resource_group_name:
        namespace.vnet_resource_group_name = namespace.resource_group_name


def validate_worker_count(namespace):
    if namespace.worker_count:
        if namespace.worker_count < 3:
            raise InvalidArgumentValueError('--worker-count must be greater than or equal to 3.')


def validate_worker_vm_disk_size_gb(namespace):
    if namespace.worker_vm_disk_size_gb:
        if namespace.worker_vm_disk_size_gb < 128:
            raise InvalidArgumentValueError('--worker-vm-disk-size-gb must be greater than or equal to 128.')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from knack.prompting import prompt_pass, NoTTYException
from knack.util import CLIError
from azure.cli.core.util import sdk_no_wait, user_confirmation

logger = get_logger(__name__)


# pylint: disable=too-many-locals, too-many-branches, too-many-statements, unused-argument
def create_cluster(cmd, client, cluster_name, resource_group_name, cluster_type,
                   location=None, tags=None, no_wait=False, cluster_version='default', cluster_tier=None,
                   cluster_configurations=None, component_version=None,
                   headnode_size='large', workernode_size='large', zookeepernode_size=None, edgenode_size=None,
                   kafka_management_node_size=None, kafka_management_node_count=2,
                   kafka_client_group_id=None, kafka_client_group_name=None,
                   workernode_count=3, workernode_data_disks_per_node=None,
                   workernode_data_disk_storage_account_type=None, workernode_data_disk_size=None,
                   http_username=None, http_password=None,
                   ssh_username='sshuser', ssh_password=None, ssh_public_key=None,
                   storage_account=None, storage_account_key=None,
                   storage_default_container=None, storage_default_filesystem=None,
                   storage_account_managed_identity=None,
                   vnet_name=None, subnet=None,
                   domain=None, ldaps_urls=None,
                   cluster_admin_account=None, cluster_admin_password=None,
                   cluster_users_group_dns=None,
                   assign_identity=None,
                   minimal_tls_version=None,
                   encryption_vault_uri=None, encryption_key_name=None, encryption_key_version=None,
                   encryption_algorithm='RSA-OAEP', encryption_in_transit=None,
                   autoscale_type=None, autoscale_min_workernode_count=None, autoscale_max_workernode_count=None,
                   timezone=None, days=None, time=None, autoscale_workernode_count=None,
                   encryption_at_host=None, esp=False, idbroker=False,
                   resource_provider_connection=None, enable_private_link=None, enable_compute_isolation=None,
                   host_sku=None, no_validation_timeout=False):
    from .util import build_identities_info, build_virtual_network_profile, parse_domain_name, \
        get_storage_account_endpoint, validate_esp_cluster_create_params
    from azure.mgmt.hdinsight.models import ClusterCreateParametersExtended, ClusterCreateProperties, OSType, \
        ClusterDefinition, ComputeProfile, HardwareProfile, Role, OsProfile, LinuxOperatingSystemProfile, \
        StorageProfile, StorageAccount, DataDisksGroups, SecurityProfile, \
        DirectoryType, DiskEncryptionProperties, Tier, SshProfile, SshPublicKey, \
        KafkaRestProperties, ClientGroupInfo, EncryptionInTransitProperties, \
        Autoscale, AutoscaleCapacity, AutoscaleRecurrence, AutoscaleSchedule, AutoscaleTimeAndCapacity, \
        NetworkProperties, PrivateLink, ComputeIsolationProperties

    validate_esp_cluster_create_params(esp, cluster_name, resource_group_name, cluster_type,
                                       subnet, domain, cluster_admin_account, assign_identity,
                                       ldaps_urls, cluster_admin_password, cluster_users_group_dns)

    if esp:
        if cluster_tier == Tier.standard:
            raise CLIError('Cluster tier cannot be {} when --esp is specified. '
                           'Please use default value or specify {} explicitly.'.format(Tier.standard, Tier.premium))
        if not cluster_tier:
            cluster_tier = Tier.premium

    # Update optional parameters with defaults
    location = location or _get_rg_location(cmd.cli_ctx, resource_group_name)

    # Format dictionary/free-form arguments
    if not cluster_configurations:
        cluster_configurations = dict()

    if component_version:
        # See validator
        component_version = {c: v for c, v in [version.split('=') for version in component_version]}

    # Validate whether HTTP credentials were provided
    if 'gateway' in cluster_configurations:
        gateway_config = cluster_configurations['gateway']
    else:
        gateway_config = dict()
    if http_username and 'restAuthCredential.username' in gateway_config:
        raise CLIError('An HTTP username must be specified either as a command-line parameter '
                       'or in the cluster configuration, but not both.')
    if not http_username:
        http_username = 'admin'  # Implement default logic here, in case a user specifies the username in configurations

    if not http_password:
        try:
            http_password = prompt_pass('HTTP password for the cluster:', confirm=True)
        except NoTTYException:
            raise CLIError('Please specify --http-password in non-interactive mode.')

    # Update the cluster config with the HTTP credentials
    gateway_config['restAuthCredential.isEnabled'] = 'true'  # HTTP credentials are required
    http_username = http_username or gateway_config['restAuthCredential.username']
    gateway_config['restAuthCredential.username'] = http_username
    gateway_config['restAuthCredential.password'] = http_password
    cluster_configurations['gateway'] = gateway_config

    # Validate whether SSH credentials were provided
    if not (ssh_password or ssh_public_key):
        logger.warning("SSH credentials not specified. Using the HTTP password as the SSH password.")
        ssh_password = http_password

    # Validate storage arguments from the user
    if storage_default_container and storage_default_filesystem:
        raise CLIError('Either the default container or the default filesystem can be specified, but not both.')

    # Retrieve primary blob service endpoint
    is_wasb = not storage_account_managed_identity
    storage_account_endpoint = None
    if storage_account:
        storage_account_endpoint = get_storage_account_endpoint(cmd, storage_account, is_wasb)

    # Attempt to infer the storage account key from the endpoint
    if not storage_account_key and storage_account and is_wasb:
        from .util import get_key_for_storage_account
        logger.info('Storage account key not specified. Attempting to retrieve key...')
        key = get_key_for_storage_account(cmd, storage_account)
        if not key:
            raise CLIError('Storage account key could not be inferred from storage account.')
        storage_account_key = key

    # Attempt to provide a default container for WASB storage accounts
    if not storage_default_container and storage_account and is_wasb:
        storage_default_container = cluster_name.lower()
        logger.warning('Default WASB container not specified, using "%s".', storage_default_container)
    elif not storage_default_filesystem and not is_wasb:
        storage_default_filesystem = cluster_name.lower()
        logger.warning('Default ADLS file system not specified, using "%s".', storage_default_filesystem)

    # Validate storage info parameters
    if is_wasb and not _all_or_none(storage_account, storage_account_key, storage_default_container):
        raise CLIError('If storage details are specified, the storage account, storage account key, '
                       'and the default container must be specified.')
    if not is_wasb and not _all_or_none(storage_account, storage_default_filesystem):
        raise CLIError('If storage details are specified, the storage account, '
                       'and the default filesystem must be specified.')

    # Validate disk encryption parameters
    if not _all_or_none(encryption_vault_uri, encryption_key_name, encryption_key_version):
        raise CLIError('Either the encryption vault URI, key name and key version should be specified, '
                       'or none of them should be.')

    # Validate kafka rest proxy parameters
    if not _all_or_none(kafka_client_group_id, kafka_client_group_name):
        raise CLIError('Either the kafka client group id and kafka client group name should be specified, '
                       'or none of them should be')

    # Validate and initialize autoscale setting
    autoscale_configuration = None
    load_based_type = "Load"
    schedule_based_type = "Schedule"
    if autoscale_type and autoscale_type.lower() == load_based_type.lower():
        if not all([autoscale_min_workernode_count, autoscale_max_workernode_count]):
            raise CLIError(
                'When the --autoscale-type is Load, '
                'both --autoscale-min-workernode-count and --autoscale-max-workernode-count should be specified.')

        autoscale_configuration = Autoscale(
            capacity=AutoscaleCapacity(
                min_instance_count=autoscale_min_workernode_count,
                max_instance_count=autoscale_max_workernode_count
            )
        )
    elif autoscale_type and autoscale_type.lower() == schedule_based_type.lower():
        if not all([timezone, days, time, autoscale_workernode_count]):
            raise CLIError(
                'When the --autoscale-type is Schedule, all of the --timezone, --days, --time, '
                '--autoscale-workernode-count should be specified.')

        autoscale_configuration = Autoscale(
            recurrence=AutoscaleRecurrence(
                time_zone=timezone,
                schedule=[AutoscaleSchedule(
                    days=days,
                    time_and_capacity=AutoscaleTimeAndCapacity(
                        time=time,
                        min_instance_count=autoscale_workernode_count,
                        max_instance_count=autoscale_workernode_count
                    )
                )]
            )
        )

    # Specify virtual network profile only when network arguments are provided
    virtual_network_profile = subnet and build_virtual_network_profile(subnet)

    # Validate data disk parameters
    if not workernode_data_disks_per_node and workernode_data_disk_storage_account_type:
        raise CLIError("Cannot define data disk storage account type unless disks per node is defined.")
    if not workernode_data_disks_per_node and workernode_data_disk_size:
        raise CLIError("Cannot define data disk size unless disks per node is defined.")
    # Specify data disk groups only when disk arguments are provided
    workernode_data_disk_groups = workernode_data_disks_per_node and [
        DataDisksGroups(
            disks_per_node=workernode_data_disks_per_node,
            storage_account_type=workernode_data_disk_storage_account_type,
            disk_size_gb=workernode_data_disk_size
        )
    ]

    os_profile = OsProfile(
        linux_operating_system_profile=LinuxOperatingSystemProfile(
            username=ssh_username,
            password=ssh_password,
            ssh_profile=ssh_public_key and SshProfile(
                public_keys=[SshPublicKey(
                    certificate_data=ssh_public_key
                )]
            )
        )
    )

    roles = [
        # Required roles
        Role(
            name="headnode",
            target_instance_count=2,
            hardware_profile=HardwareProfile(vm_size=headnode_size),
            os_profile=os_profile,
            virtual_network_profile=virtual_network_profile
        ),
        Role(
            name="workernode",
            target_instance_count=workernode_count,
            hardware_profile=HardwareProfile(vm_size=workernode_size),
            os_profile=os_profile,
            virtual_network_profile=virtual_network_profile,
            data_disks_groups=workernode_data_disk_groups,
            autoscale_configuration=autoscale_configuration
        )
    ]

    if zookeepernode_size:
        roles.append(
            Role(
                name="zookeepernode",
                target_instance_count=3,
                hardware_profile=HardwareProfile(vm_size=zookeepernode_size),
                os_profile=os_profile,
                virtual_network_profile=virtual_network_profile
            ))
    if edgenode_size:
        roles.append(
            Role(
                name="edgenode",
                target_instance_count=1,
                hardware_profile=HardwareProfile(vm_size=edgenode_size),
                os_profile=os_profile,
                virtual_network_profile=virtual_network_profile
            ))
    if kafka_management_node_size:
        # generate kafkaRestProperties
        roles.append(
            Role(
                name="kafkamanagementnode",
                target_instance_count=kafka_management_node_count,
                hardware_profile=HardwareProfile(vm_size=kafka_management_node_size),
                os_profile=os_profile,
                virtual_network_profile=virtual_network_profile
            )
        )

    if esp and idbroker:
        roles.append(
            Role(
                name="idbrokernode",
                target_instance_count=2,
                virtual_network_profile=virtual_network_profile
            )
        )

    storage_accounts = []
    if storage_account:
        # Specify storage account details only when storage arguments are provided
        storage_accounts.append(
            StorageAccount(
                name=storage_account_endpoint,
                key=storage_account_key,
                container=storage_default_container,
                file_system=storage_default_filesystem,
                resource_id=storage_account,
                msi_resource_id=storage_account_managed_identity,
                is_default=True
            )
        )

    additional_storage_accounts = []  # TODO: Add support for additional storage accounts
    if additional_storage_accounts:
        storage_accounts += [
            StorageAccount(
                name=s.storage_account_endpoint,
                key=s.storage_account_key,
                container=s.container,
                is_default=False
            )
            for s in additional_storage_accounts
        ]

    assign_identities = []
    if assign_identity:
        assign_identities.append(assign_identity)

    if storage_account_managed_identity:
        assign_identities.append(storage_account_managed_identity)

    cluster_identity = build_identities_info(assign_identities) if assign_identities else None

    domain_name = domain and parse_domain_name(domain)
    if not ldaps_urls and domain_name:
        ldaps_urls = ['ldaps://{}:636'.format(domain_name)]

    security_profile = domain and SecurityProfile(
        directory_type=DirectoryType.active_directory,
        domain=domain_name,
        ldaps_urls=ldaps_urls,
        domain_username=cluster_admin_account,
        domain_user_password=cluster_admin_password,
        cluster_users_group_dns=cluster_users_group_dns,
        aadds_resource_id=domain,
        msi_resource_id=assign_identity
    )

    disk_encryption_properties = encryption_vault_uri and DiskEncryptionProperties(
        vault_uri=encryption_vault_uri,
        key_name=encryption_key_name,
        key_version=encryption_key_version,
        encryption_algorithm=encryption_algorithm,
        msi_resource_id=assign_identity
    )

    if encryption_at_host:
        if disk_encryption_properties:
            disk_encryption_properties.encryption_at_host = encryption_at_host
        else:
            disk_encryption_properties = DiskEncryptionProperties(encryption_at_host=encryption_at_host)

    kafka_rest_properties = (kafka_client_group_id and kafka_client_group_name) and KafkaRestProperties(
        client_group_info=ClientGroupInfo(
            group_id=kafka_client_group_id,
            group_name=kafka_client_group_name
        )
    )

    encryption_in_transit_properties = encryption_in_transit and EncryptionInTransitProperties(
        is_encryption_in_transit_enabled=encryption_in_transit
    )

    # relay outbound and private link
    network_properties = (resource_provider_connection or enable_private_link) and NetworkProperties(
        resource_provider_connection=resource_provider_connection,
        private_link=PrivateLink.enabled if enable_private_link is True else PrivateLink.disabled
    )

    # compute isolation
    compute_isolation_properties = enable_compute_isolation and ComputeIsolationProperties(
        enable_compute_isolation=enable_compute_isolation,
        host_sku=host_sku
    )

    create_params = ClusterCreateParametersExtended(
        location=location,
        tags=tags,
        properties=ClusterCreateProperties(
            cluster_version=cluster_version,
            os_type=OSType.linux,
            tier=cluster_tier,
            cluster_definition=ClusterDefinition(
                kind=cluster_type,
                configurations=cluster_configurations,
                component_version=component_version
            ),
            compute_profile=ComputeProfile(
                roles=roles
            ),
            storage_profile=StorageProfile(
                storageaccounts=storage_accounts
            ),
            security_profile=security_profile,
            disk_encryption_properties=disk_encryption_properties,
            kafka_rest_properties=kafka_rest_properties,
            min_supported_tls_version=minimal_tls_version,
            encryption_in_transit_properties=encryption_in_transit_properties,
            network_properties=network_properties,
            compute_isolation_properties=compute_isolation_properties
        ),
        identity=cluster_identity
    )

    if no_wait:
        return sdk_no_wait(no_wait, client.create, resource_group_name, cluster_name, create_params)

    return client.create(resource_group_name, cluster_name, create_params)


def list_clusters(cmd, client, resource_group_name=None):  # pylint: disable=unused-argument
    clusters_list = client.list_by_resource_group(resource_group_name=resource_group_name) \
        if resource_group_name else client.list()

    return list(clusters_list)


# pylint: disable=unused-argument
def rotate_hdi_cluster_key(cmd, client, resource_group_name, cluster_name,
                           encryption_vault_uri, encryption_key_name, encryption_key_version, no_wait=False):
    from azure.mgmt.hdinsight.models import ClusterDiskEncryptionParameters
    rotate_params = ClusterDiskEncryptionParameters(
        vault_uri=encryption_vault_uri,
        key_name=encryption_key_name,
        key_version=encryption_key_version
    )

    if no_wait:
        return sdk_no_wait(no_wait, client.rotate_disk_encryption_key, resource_group_name, cluster_name, rotate_params)

    return client.rotate_disk_encryption_key(resource_group_name, cluster_name, rotate_params)


def _all_or_none(*params):
    return all(params) or not any(params)


def _get_rg_location(ctx, resource_group_name, subscription_id=None):
    from ._client_factory import cf_resource_groups
    groups = cf_resource_groups(ctx, subscription_id=subscription_id)
    # Just do the get, we don't need the result, it will error out if the group doesn't exist.
    rg = groups.get(resource_group_name)
    return rg.location


# pylint: disable=unused-argument
def create_hdi_application(cmd, client, resource_group_name, cluster_name, application_name,
                           script_uri, script_action_name, script_parameters=None, edgenode_size='Standard_D3_V2',
                           ssh_username='sshuser', ssh_password=None, ssh_public_key=None,
                           marketplace_identifier=None, application_type='CustomApplication', tags=None,
                           https_endpoint_access_mode='WebPage', https_endpoint_destination_port=8080,
                           sub_domain_suffix=None, disable_gateway_auth=None,
                           vnet_name=None, subnet=None, no_validation_timeout=False):
    from .util import build_virtual_network_profile
    from azure.mgmt.hdinsight.models import Application, ApplicationProperties, ComputeProfile, RuntimeScriptAction, \
        Role, LinuxOperatingSystemProfile, HardwareProfile, \
        ApplicationGetHttpsEndpoint, OsProfile, SshProfile, SshPublicKey

    # Specify virtual network profile only when network arguments are provided
    virtual_network_profile = subnet and build_virtual_network_profile(subnet)

    os_profile = (ssh_password or ssh_public_key) and OsProfile(
        linux_operating_system_profile=LinuxOperatingSystemProfile(
            username=ssh_username,
            password=ssh_password,
            ssh_profile=ssh_public_key and SshProfile(
                public_keys=[SshPublicKey(
                    certificate_data=ssh_public_key
                )]
            )
        )
    )

    roles = [
        Role(
            name="edgenode",
            target_instance_count=1,
            hardware_profile=HardwareProfile(vm_size=edgenode_size),
            os_profile=os_profile,
            virtual_network_profile=virtual_network_profile
        )
    ]

    # Validate network profile parameters
    https_endpoints = []
    if sub_domain_suffix:
        https_endpoints.append(
            ApplicationGetHttpsEndpoint(
                access_modes=[https_endpoint_access_mode],
                destination_port=https_endpoint_destination_port,
                sub_domain_suffix=sub_domain_suffix,
                disable_gateway_auth=disable_gateway_auth
            )
        )

    application_properties = ApplicationProperties(
        compute_profile=ComputeProfile(
            roles=roles
        ),
        install_script_actions=[
            RuntimeScriptAction(
                name=script_action_name,
                uri=script_uri,
                parameters=script_parameters,
                roles=[role.name for role in roles]
            )
        ],
        https_endpoints=https_endpoints,
        application_type=application_type,
        marketplace_identifier=marketplace_identifier,
    )

    create_params = Application(
        tags=tags,
        properties=application_properties
    )

    return client.create(resource_group_name, cluster_name, application_name, create_params)


# pylint: disable=unused-argument
def enable_hdi_monitoring(cmd, client, resource_group_name, cluster_name, workspace,
                          primary_key=None, workspace_type='resource_id', no_validation_timeout=False):
    from msrestazure.tools import parse_resource_id
    from ._client_factory import cf_log_analytics

    if workspace_type != 'resource_id' and not primary_key:
        raise CLIError('primary key is required when workspace ID is provided')

    workspace_id = workspace
    if workspace_type == 'resource_id':
        parsed_workspace = parse_resource_id(workspace)
        workspace_resource_group_name = parsed_workspace['resource_group']
        workspace_name = parsed_workspace['resource_name']

        log_analytics_client = cf_log_analytics(cmd.cli_ctx)
        log_analytics_workspace = log_analytics_client.workspaces.get(workspace_resource_group_name, workspace_name)
        if not log_analytics_workspace:
            raise CLIError('Fails to retrieve workspace by {}'.format(workspace))

        # Only retrieve primary key when not provided
        if not primary_key:
            shared_keys = log_analytics_client.shared_keys.get_shared_keys(workspace_resource_group_name,
                                                                           workspace_name)
            if not shared_keys:
                raise CLIError('Fails to retrieve shared key for workspace {}'.format(log_analytics_workspace))

            primary_key = shared_keys.primary_shared_key

        workspace_id = log_analytics_workspace.customer_id

    return client.enable_monitoring(
        resource_group_name,
        cluster_name,
        workspace_id,
        primary_key)


# pylint: disable=unused-argument
def execute_hdi_script_action(cmd, client, resource_group_name, cluster_name,
                              script_uri, script_action_name, roles, script_parameters=None,
                              persist_on_success=False):
    from azure.mgmt.hdinsight.models import RuntimeScriptAction

    script_actions_params = [
        RuntimeScriptAction(
            name=script_action_name,
            uri=script_uri,
            parameters=script_parameters,
            roles=roles
        )
    ]

    return client.execute_script_actions(resource_group_name, cluster_name, persist_on_success, script_actions_params)


# pylint: disable=redefined-builtin
def create_autoscale(cmd, client, resource_group_name, cluster_name, type, min_workernode_count=None,
                     max_workernode_count=None, timezone=None, days=None, time=None, workernode_count=None,
                     no_wait=False, yes=False):
    from azure.mgmt.hdinsight.models import Autoscale, AutoscaleCapacity, AutoscaleRecurrence, AutoscaleSchedule, \
        AutoscaleTimeAndCapacity
    load_based_type = "Load"
    schedule_based_type = "Schedule"
    autoscale_configuration = None

    if type.lower() == load_based_type.lower():
        if not all([min_workernode_count, max_workernode_count]):
            raise CLIError(
                'When the --type is Load, both --min-workernode-count and --max-workernode-count should be specified.')

        autoscale_configuration = Autoscale(
            capacity=AutoscaleCapacity(
                min_instance_count=min_workernode_count,
                max_instance_count=max_workernode_count
            )
        )
    elif type.lower() == schedule_based_type.lower():
        if not all([timezone, days, time, workernode_count]):
            raise CLIError(
                'When the --type is Schedule, all of the --timezone, --days, --time, '
                '--workernode-count should be specified.')

        autoscale_configuration = Autoscale(
            recurrence=AutoscaleRecurrence(
                time_zone=timezone,
                schedule=[AutoscaleSchedule(
                    days=days,
                    time_and_capacity=AutoscaleTimeAndCapacity(
                        time=time,
                        min_instance_count=workernode_count,
                        max_instance_count=workernode_count
                    )
                )]
            )
        )

    user_confirmation(
        "This operation will override previous autoscale configuration. "
        "Are you sure you want to perform this operation", yes)

    return sdk_no_wait(no_wait, client.update_auto_scale_configuration, resource_group_name, cluster_name,
                       autoscale_configuration)


def update_autoscale(cmd, client, resource_group_name, cluster_name, min_workernode_count=None,
                     max_workernode_count=None, timezone=None, no_wait=False):
    from azure.mgmt.hdinsight.models import AutoscaleCapacity, AutoscaleRecurrence
    cluster = client.get(resource_group_name, cluster_name)
    autoscale_configuration = _extract_and_validate_autoscale_configuration(cluster)

    # try to update load-based configuration
    if autoscale_configuration.capacity:
        if min_workernode_count:
            autoscale_configuration.capacity.min_instance_count = min_workernode_count
        if max_workernode_count:
            autoscale_configuration.capacity.max_instance_count = max_workernode_count

    if not autoscale_configuration.capacity:
        if min_workernode_count or max_workernode_count:
            autoscale_configuration.capacity = AutoscaleCapacity()
            if min_workernode_count:
                autoscale_configuration.capacity.min_instance_count = min_workernode_count
            if max_workernode_count:
                autoscale_configuration.capacity.max_instance_count = max_workernode_count

    # try to update schedule-based configuration
    if timezone:
        if autoscale_configuration.recurrence:
            autoscale_configuration.recurrence.time_zone = timezone
        elif not autoscale_configuration.recurrence:
            autoscale_configuration.recurrence = AutoscaleRecurrence(time_zone=timezone)

    return sdk_no_wait(no_wait, client.update_auto_scale_configuration, resource_group_name, cluster_name,
                       autoscale_configuration)


def show_autoscale(cmd, client, resource_group_name, cluster_name):
    cluster = client.get(resource_group_name, cluster_name)
    autoscale_configuration = _extract_and_validate_autoscale_configuration(cluster)
    return autoscale_configuration


def delete_autoscale(cmd, client, resource_group_name, cluster_name, no_wait=False):
    cluster = client.get(resource_group_name, cluster_name)
    _extract_and_validate_autoscale_configuration(cluster)
    autoscale_configuration = None

    return sdk_no_wait(no_wait, client.update_auto_scale_configuration, resource_group_name, cluster_name,
                       autoscale_configuration)


def list_timezones(cmd, client):
    from .util import AUTOSCALE_TIMEZONES
    return AUTOSCALE_TIMEZONES


def create_autoscale_condition(cmd, client, resource_group_name, cluster_name, days, time, workernode_count,
                               no_wait=False):
    from azure.mgmt.hdinsight.models import AutoscaleRecurrence, AutoscaleSchedule, AutoscaleTimeAndCapacity
    cluster = client.get(resource_group_name, cluster_name)
    autoscale_configuration = _extract_and_validate_autoscale_configuration(cluster)

    # try to add schedule condition
    condition = AutoscaleSchedule(
        days=days,
        time_and_capacity=AutoscaleTimeAndCapacity(
            time=time,
            min_instance_count=workernode_count,
            max_instance_count=workernode_count)
    )
    if autoscale_configuration.recurrence:
        autoscale_configuration.recurrence.schedule.append(condition)
    elif not autoscale_configuration.recurrence:
        autoscale_configuration.recurrence = AutoscaleRecurrence(
            schedule=[condition]
        )

    return sdk_no_wait(no_wait, client.update_auto_scale_configuration, resource_group_name, cluster_name,
                       autoscale_configuration)


def update_autoscale_condition(cmd, client, resource_group_name, cluster_name, index, days=None, time=None,
                               workernode_count=None, no_wait=False):
    cluster = client.get(resource_group_name, cluster_name)
    autoscale_configuration = _extract_and_validate_autoscale_configuration(cluster)
    _validate_schedule_configuration(autoscale_configuration)

    if index >= len(autoscale_configuration.recurrence.schedule):
        raise CLIError(
            'This cluster only has {} condition(s). Please specify a correct index which starts with 0.'.format(
                len(autoscale_configuration.recurrence.schedule)))

    # try to update schedule condition
    if days:
        autoscale_configuration.recurrence.schedule[index].days = days
    if time:
        autoscale_configuration.recurrence.schedule[index].time_and_capacity.time = time
    if workernode_count:
        autoscale_configuration.recurrence.schedule[index].time_and_capacity.min_instance_count = workernode_count
        autoscale_configuration.recurrence.schedule[index].time_and_capacity.max_instance_count = workernode_count

    return sdk_no_wait(no_wait, client.update_auto_scale_configuration, resource_group_name, cluster_name,
                       autoscale_configuration)


def delete_autoscale_condition(cmd, client, resource_group_name, cluster_name, index, no_wait=False):
    cluster = client.get(resource_group_name, cluster_name)
    autoscale_configuration = _extract_and_validate_autoscale_configuration(cluster)
    _validate_schedule_configuration(autoscale_configuration)

    conditions_count = len(autoscale_configuration.recurrence.schedule)
    if len(index) >= conditions_count:
        raise CLIError('Deleting all conditions is not allowed.'
                       'If you want to disable autoscale please use `az hdinsight autoscale delete`.')
    autoscale_configuration.recurrence.schedule = [autoscale_configuration.recurrence.schedule[i] for i in
                                                   range(conditions_count) if i not in index]
    return sdk_no_wait(no_wait, client.update_auto_scale_configuration, resource_group_name, cluster_name,
                       autoscale_configuration)


def list_autoscale_condition(cmd, client, resource_group_name, cluster_name):
    cluster = client.get(resource_group_name, cluster_name)
    autoscale_configuration = _extract_and_validate_autoscale_configuration(cluster)
    _validate_schedule_configuration(autoscale_configuration)
    return autoscale_configuration.recurrence.schedule


def _extract_and_validate_autoscale_configuration(cluster):
    autoscale_configuration = None
    for role in cluster.properties.compute_profile.roles:
        if role.name == "workernode":
            autoscale_configuration = role.autoscale_configuration
            break
    if not autoscale_configuration:
        raise CLIError('The cluster has not enabled autoscale.')
    return autoscale_configuration


def _validate_schedule_configuration(autoscale_configuration):
    if not autoscale_configuration.recurrence:
        raise CLIError('The cluster has not enabled Schedule-based autoscale.')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from distutils.version import StrictVersion  # pylint: disable=no-name-in-module,import-error
# pylint: disable=no-name-in-module,import-error
from azure.mgmt.containerservice.v2021_02_01.models import ManagedClusterAPIServerAccessProfile
from knack.util import CLIError
from ._consts import CONST_OUTBOUND_TYPE_LOAD_BALANCER, CONST_OUTBOUND_TYPE_USER_DEFINED_ROUTING


def _populate_api_server_access_profile(api_server_authorized_ip_ranges, enable_private_cluster=False, instance=None):
    if instance is None or instance.api_server_access_profile is None:
        profile = ManagedClusterAPIServerAccessProfile()
    else:
        profile = instance.api_server_access_profile

    if enable_private_cluster:
        profile.enable_private_cluster = True

    if api_server_authorized_ip_ranges is None or api_server_authorized_ip_ranges == "":
        authorized_ip_ranges = []
    else:
        authorized_ip_ranges = [ip.strip() for ip in api_server_authorized_ip_ranges.split(",")]

    if profile.enable_private_cluster and authorized_ip_ranges:
        raise CLIError('--api-server-authorized-ip-ranges is not supported for private cluster')

    profile.authorized_ip_ranges = authorized_ip_ranges
    return profile


def _set_load_balancer_sku(load_balancer_sku, kubernetes_version):
    if load_balancer_sku:
        return load_balancer_sku
    if kubernetes_version and StrictVersion(kubernetes_version) < StrictVersion("1.13.0"):
        print('Setting load_balancer_sku to basic as it is not specified and kubernetes \
        version(%s) less than 1.13.0 only supports basic load balancer SKU\n' % (kubernetes_version))
        return "basic"
    return "standard"


def _set_vm_set_type(vm_set_type, kubernetes_version):
    if not vm_set_type:
        if kubernetes_version and StrictVersion(kubernetes_version) < StrictVersion("1.12.9"):
            print('Setting vm_set_type to availabilityset as it is \
            not specified and kubernetes version(%s) less than 1.12.9 only supports \
            availabilityset\n' % (kubernetes_version))
            vm_set_type = "AvailabilitySet"

    if not vm_set_type:
        vm_set_type = "VirtualMachineScaleSets"

    # normalize as server validation is case-sensitive
    if vm_set_type.lower() == "AvailabilitySet".lower():
        vm_set_type = "AvailabilitySet"

    if vm_set_type.lower() == "VirtualMachineScaleSets".lower():
        vm_set_type = "VirtualMachineScaleSets"
    return vm_set_type


def _set_outbound_type(outbound_type, vnet_subnet_id, load_balancer_sku, load_balancer_profile):
    if outbound_type != CONST_OUTBOUND_TYPE_USER_DEFINED_ROUTING:
        return CONST_OUTBOUND_TYPE_LOAD_BALANCER

    if load_balancer_sku == "basic":
        raise CLIError("userDefinedRouting doesn't support basic load balancer sku")

    if vnet_subnet_id in ["", None]:
        raise CLIError("--vnet-subnet-id must be specified for userDefinedRouting and it must \
        be pre-configured with a route table with egress rules")

    if load_balancer_profile:
        if (load_balancer_profile.managed_outbound_ips or
                load_balancer_profile.outbound_ips or
                load_balancer_profile.outbound_ip_prefixes):
            raise CLIError("userDefinedRouting doesn't support customizing a standard load balancer with IP addresses")

    return CONST_OUTBOUND_TYPE_USER_DEFINED_ROUTING


def _parse_comma_separated_list(text):
    if text is None:
        return None
    if text == "":
        return []
    return text.split(",")



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from knack.log import get_logger

logger = get_logger(__name__)


# region Private Link Scope
def show_private_link_scope(client, resource_group_name, scope_name):
    return client.get(resource_group_name=resource_group_name,
                      scope_name=scope_name)


def delete_private_link_scope(client, resource_group_name, scope_name):
    return client.begin_delete(resource_group_name=resource_group_name,
                               scope_name=scope_name)


def list_private_link_scope(client, resource_group_name=None):
    if not resource_group_name:
        return client.list()
    return client.list_by_resource_group(resource_group_name=resource_group_name)


def create_private_link_scope(client, resource_group_name, scope_name, location='Global', tags=None):
    from azure.mgmt.monitor.models import AzureMonitorPrivateLinkScope
    private_link_scope = AzureMonitorPrivateLinkScope(location=location, tags=tags)
    return client.create_or_update(resource_group_name=resource_group_name,
                                   scope_name=scope_name,
                                   azure_monitor_private_link_scope_payload=private_link_scope)


def update_private_link_scope(client, resource_group_name, scope_name, tags):
    from azure.mgmt.monitor.models import TagsResource
    tags = TagsResource(tags=tags)
    return client.update_tags(resource_group_name=resource_group_name,
                              scope_name=scope_name,
                              private_link_scope_tags=tags)
# endregion


# region Private Link Scope Resource
def show_private_link_scope_resource(client, resource_group_name, scope_name, resource_name):
    return client.get(resource_group_name=resource_group_name,
                      scope_name=scope_name,
                      name=resource_name)


def delete_private_link_scope_resource(client, resource_group_name, scope_name, resource_name):
    return client.begin_delete(resource_group_name=resource_group_name,
                               scope_name=scope_name,
                               name=resource_name)


def list_private_link_scope_resource(client, resource_group_name, scope_name):
    return client.list_by_private_link_scope(resource_group_name=resource_group_name,
                                             scope_name=scope_name)


def create_private_link_scope_resource(client, resource_group_name, scope_name, resource_name,
                                       linked_resource_id):
    from azure.mgmt.monitor.models import ScopedResource
    scoped_resource = ScopedResource(linked_resource_id=linked_resource_id)
    return client.begin_create_or_update(resource_group_name=resource_group_name,
                                         scope_name=scope_name,
                                         name=resource_name,
                                         parameters=scoped_resource)

# endregion


# region Private Link Resource
def list_private_link_resource(client, resource_group_name, scope_name):
    return client.list_by_private_link_scope(resource_group_name=resource_group_name,
                                             scope_name=scope_name)


def show_private_link_resource(client, resource_group_name, scope_name, group_name):
    return client.get(resource_group_name=resource_group_name,
                      scope_name=scope_name,
                      group_name=group_name)
# endregion


# region Private Endpoint Connection
def show_private_endpoint_connection(client, resource_group_name, scope_name, private_endpoint_connection_name):
    return client.get(resource_group_name=resource_group_name,
                      scope_name=scope_name,
                      private_endpoint_connection_name=private_endpoint_connection_name)


def delete_private_endpoint_connection(client, resource_group_name, scope_name, private_endpoint_connection_name):
    return client.begin_delete(resource_group_name=resource_group_name,
                               scope_name=scope_name,
                               private_endpoint_connection_name=private_endpoint_connection_name)


def list_private_endpoint_connection(client, resource_group_name, scope_name):
    return client.list_by_private_link_scope(resource_group_name=resource_group_name,
                                             scope_name=scope_name)


# pylint: disable=line-too-long, unused-argument
def _update_private_endpoint_connection_status(cmd, client, resource_group_name, scope_name,
                                               private_endpoint_connection_name, is_approved=True, description=None):
    private_endpoint_connection = client.get(resource_group_name=resource_group_name, scope_name=scope_name,
                                             private_endpoint_connection_name=private_endpoint_connection_name)

    old_status = private_endpoint_connection.private_link_service_connection_state.status
    new_status = "Approved" if is_approved else "Rejected"
    if old_status == new_status:
        logger.warning('The status has been satisfied. Skip this command.')
        return None
    private_endpoint_connection.private_link_service_connection_state.status = new_status
    private_endpoint_connection.private_link_service_connection_state.description = description
    from azure.mgmt.monitor.models import PrivateEndpointConnection
    private_endpoint_connection = PrivateEndpointConnection(private_link_service_connection_state=private_endpoint_connection.private_link_service_connection_state)
    return client.begin_create_or_update(resource_group_name=resource_group_name,
                                         scope_name=scope_name,
                                         private_endpoint_connection_name=private_endpoint_connection_name,
                                         parameters=private_endpoint_connection)


def approve_private_endpoint_connection(cmd, client, resource_group_name, scope_name,
                                        private_endpoint_connection_name, description=""):

    return _update_private_endpoint_connection_status(
        cmd, client, resource_group_name=resource_group_name, scope_name=scope_name, is_approved=True,
        private_endpoint_connection_name=private_endpoint_connection_name, description=description
    )


def reject_private_endpoint_connection(cmd, client, resource_group_name, scope_name,
                                       private_endpoint_connection_name, description=""):
    return _update_private_endpoint_connection_status(
        cmd, client, resource_group_name=resource_group_name, scope_name=scope_name, is_approved=False,
        private_endpoint_connection_name=private_endpoint_connection_name, description=description
    )
# endregion



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# Generated from MetricAlertCondition.g4 by ANTLR 4.7.2
# encoding: utf-8
# pylint: disable=all
from antlr4 import *
from io import StringIO
import sys


def serializedATN():
    with StringIO() as buf:
        buf.write(u"\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3")
        buf.write(u"\34\u00b5\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7")
        buf.write(u"\t\7\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t")
        buf.write(u"\r\4\16\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22")
        buf.write(u"\4\23\t\23\4\24\t\24\4\25\t\25\4\26\t\26\4\27\t\27\4")
        buf.write(u"\30\t\30\3\2\3\2\3\2\3\2\7\2\65\n\2\f\2\16\28\13\2\3")
        buf.write(u"\2\3\2\3\2\3\2\3\2\3\2\5\2@\n\2\3\2\3\2\3\2\5\2E\n\2")
        buf.write(u"\3\2\3\2\7\2I\n\2\f\2\16\2L\13\2\3\2\7\2O\n\2\f\2\16")
        buf.write(u"\2R\13\2\3\3\3\3\3\3\3\4\6\4X\n\4\r\4\16\4Y\3\5\6\5]")
        buf.write(u"\n\5\r\5\16\5^\3\6\3\6\3\6\3\7\3\7\3\b\3\b\3\b\3\t\3")
        buf.write(u"\t\3\t\3\t\3\t\3\t\3\t\3\t\3\t\7\tr\n\t\f\t\16\tu\13")
        buf.write(u"\t\3\n\3\n\3\n\3\13\3\13\3\13\3\f\3\f\3\f\3\r\3\r\3\16")
        buf.write(u"\3\16\3\16\3\17\6\17\u0086\n\17\r\17\16\17\u0087\3\20")
        buf.write(u"\3\20\3\20\3\21\3\21\3\21\3\21\3\21\7\21\u0092\n\21\f")
        buf.write(u"\21\16\21\u0095\13\21\3\22\3\22\3\22\3\22\3\23\3\23\3")
        buf.write(u"\23\3\24\3\24\3\24\3\25\3\25\3\25\3\26\3\26\3\26\3\27")
        buf.write(u"\3\27\3\27\3\27\7\27\u00ab\n\27\f\27\16\27\u00ae\13\27")
        buf.write(u"\3\30\6\30\u00b1\n\30\r\30\16\30\u00b2\3\30\2\2\31\2")
        buf.write(u"\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(*,.\2\t\5")
        buf.write(u"\2\3\4\30\30\34\34\6\2\3\13\30\30\32\32\34\34\b\2\3\3")
        buf.write(u"\7\7\t\t\f\f\30\30\34\34\4\2\n\n\20\20\3\2\21\22\4\2")
        buf.write(u"\n\n\23\23\t\2\3\3\5\5\7\13\r\16\30\30\32\32\34\34\2")
        buf.write(u"\u00a9\2\60\3\2\2\2\4S\3\2\2\2\6W\3\2\2\2\b\\\3\2\2\2")
        buf.write(u"\n`\3\2\2\2\fc\3\2\2\2\16e\3\2\2\2\20h\3\2\2\2\22v\3")
        buf.write(u"\2\2\2\24y\3\2\2\2\26|\3\2\2\2\30\177\3\2\2\2\32\u0081")
        buf.write(u"\3\2\2\2\34\u0085\3\2\2\2\36\u0089\3\2\2\2 \u008c\3\2")
        buf.write(u"\2\2\"\u0096\3\2\2\2$\u009a\3\2\2\2&\u009d\3\2\2\2(\u00a0")
        buf.write(u"\3\2\2\2*\u00a3\3\2\2\2,\u00a6\3\2\2\2.\u00b0\3\2\2\2")
        buf.write(u"\60\66\5\4\3\2\61\62\5\6\4\2\62\63\7\3\2\2\63\65\3\2")
        buf.write(u"\2\2\64\61\3\2\2\2\658\3\2\2\2\66\64\3\2\2\2\66\67\3")
        buf.write(u"\2\2\2\67?\3\2\2\28\66\3\2\2\29:\7\31\2\2:;\5\b\5\2;")
        buf.write(u"<\7\31\2\2<=\7\32\2\2=@\3\2\2\2>@\5\b\5\2?9\3\2\2\2?")
        buf.write(u">\3\2\2\2@A\3\2\2\2AD\5\n\6\2BE\5\f\7\2CE\5\20\t\2DB")
        buf.write(u"\3\2\2\2DC\3\2\2\2EJ\3\2\2\2FG\7\32\2\2GI\5 \21\2HF\3")
        buf.write(u"\2\2\2IL\3\2\2\2JH\3\2\2\2JK\3\2\2\2KP\3\2\2\2LJ\3\2")
        buf.write(u"\2\2MO\7\33\2\2NM\3\2\2\2OR\3\2\2\2PN\3\2\2\2PQ\3\2\2")
        buf.write(u"\2Q\3\3\2\2\2RP\3\2\2\2ST\7\34\2\2TU\7\32\2\2U\5\3\2")
        buf.write(u"\2\2VX\t\2\2\2WV\3\2\2\2XY\3\2\2\2YW\3\2\2\2YZ\3\2\2")
        buf.write(u"\2Z\7\3\2\2\2[]\t\3\2\2\\[\3\2\2\2]^\3\2\2\2^\\\3\2\2")
        buf.write(u"\2^_\3\2\2\2_\t\3\2\2\2`a\7\27\2\2ab\7\32\2\2b\13\3\2")
        buf.write(u"\2\2cd\7\30\2\2d\r\3\2\2\2ef\7\24\2\2fg\7\32\2\2g\17")
        buf.write(u"\3\2\2\2hi\5\16\b\2ij\5\22\n\2jk\5\24\13\2kl\5\26\f\2")
        buf.write(u"ls\5\30\r\2mn\7\32\2\2no\5\32\16\2op\5\34\17\2pr\3\2")
        buf.write(u"\2\2qm\3\2\2\2ru\3\2\2\2sq\3\2\2\2st\3\2\2\2t\21\3\2")
        buf.write(u"\2\2us\3\2\2\2vw\7\34\2\2wx\7\32\2\2x\23\3\2\2\2yz\7")
        buf.write(u"\30\2\2z{\7\32\2\2{\25\3\2\2\2|}\7\25\2\2}~\7\32\2\2")
        buf.write(u"~\27\3\2\2\2\177\u0080\7\30\2\2\u0080\31\3\2\2\2\u0081")
        buf.write(u"\u0082\7\26\2\2\u0082\u0083\7\32\2\2\u0083\33\3\2\2\2")
        buf.write(u"\u0084\u0086\t\4\2\2\u0085\u0084\3\2\2\2\u0086\u0087")
        buf.write(u"\3\2\2\2\u0087\u0085\3\2\2\2\u0087\u0088\3\2\2\2\u0088")
        buf.write(u"\35\3\2\2\2\u0089\u008a\7\17\2\2\u008a\u008b\7\32\2\2")
        buf.write(u"\u008b\37\3\2\2\2\u008c\u008d\5\36\20\2\u008d\u0093\5")
        buf.write(u"\"\22\2\u008e\u008f\5$\23\2\u008f\u0090\5\"\22\2\u0090")
        buf.write(u"\u0092\3\2\2\2\u0091\u008e\3\2\2\2\u0092\u0095\3\2\2")
        buf.write(u"\2\u0093\u0091\3\2\2\2\u0093\u0094\3\2\2\2\u0094!\3\2")
        buf.write(u"\2\2\u0095\u0093\3\2\2\2\u0096\u0097\5*\26\2\u0097\u0098")
        buf.write(u"\5&\24\2\u0098\u0099\5,\27\2\u0099#\3\2\2\2\u009a\u009b")
        buf.write(u"\t\5\2\2\u009b\u009c\7\32\2\2\u009c%\3\2\2\2\u009d\u009e")
        buf.write(u"\t\6\2\2\u009e\u009f\7\32\2\2\u009f\'\3\2\2\2\u00a0\u00a1")
        buf.write(u"\t\7\2\2\u00a1\u00a2\7\32\2\2\u00a2)\3\2\2\2\u00a3\u00a4")
        buf.write(u"\7\34\2\2\u00a4\u00a5\7\32\2\2\u00a5+\3\2\2\2\u00a6\u00ac")
        buf.write(u"\5.\30\2\u00a7\u00a8\5(\25\2\u00a8\u00a9\5.\30\2\u00a9")
        buf.write(u"\u00ab\3\2\2\2\u00aa\u00a7\3\2\2\2\u00ab\u00ae\3\2\2")
        buf.write(u"\2\u00ac\u00aa\3\2\2\2\u00ac\u00ad\3\2\2\2\u00ad-\3\2")
        buf.write(u"\2\2\u00ae\u00ac\3\2\2\2\u00af\u00b1\t\b\2\2\u00b0\u00af")
        buf.write(u"\3\2\2\2\u00b1\u00b2\3\2\2\2\u00b2\u00b0\3\2\2\2\u00b2")
        buf.write(u"\u00b3\3\2\2\2\u00b3/\3\2\2\2\16\66?DJPY^s\u0087\u0093")
        buf.write(u"\u00ac\u00b2")
        return buf.getvalue()


class MetricAlertConditionParser ( Parser ):

    grammarFileName = "MetricAlertCondition.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ u"<INVALID>", u"'.'", u"'/'", u"'_'", u"'\\'", u"':'", 
                     u"'%'", u"'-'", u"','", u"'|'", u"'+'", u"'*'", u"'~'" ]

    symbolicNames = [ u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"WHERE", u"AND", u"INCLUDES", u"EXCLUDES", 
                      u"OR", u"DYNAMIC", u"OF", u"SINCE", u"OPERATOR", u"NUMBER", 
                      u"QUOTE", u"WHITESPACE", u"NEWLINE", u"WORD" ]

    RULE_expression = 0
    RULE_aggregation = 1
    RULE_namespace = 2
    RULE_metric = 3
    RULE_operator = 4
    RULE_threshold = 5
    RULE_dynamic = 6
    RULE_dynamics = 7
    RULE_dyn_sensitivity = 8
    RULE_dyn_violations = 9
    RULE_dyn_of_separator = 10
    RULE_dyn_windows = 11
    RULE_dyn_since_seperator = 12
    RULE_dyn_datetime = 13
    RULE_where = 14
    RULE_dimensions = 15
    RULE_dimension = 16
    RULE_dim_separator = 17
    RULE_dim_operator = 18
    RULE_dim_val_separator = 19
    RULE_dim_name = 20
    RULE_dim_values = 21
    RULE_dim_value = 22

    ruleNames =  [ u"expression", u"aggregation", u"namespace", u"metric", 
                   u"operator", u"threshold", u"dynamic", u"dynamics", u"dyn_sensitivity", 
                   u"dyn_violations", u"dyn_of_separator", u"dyn_windows", 
                   u"dyn_since_seperator", u"dyn_datetime", u"where", u"dimensions", 
                   u"dimension", u"dim_separator", u"dim_operator", u"dim_val_separator", 
                   u"dim_name", u"dim_values", u"dim_value" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    WHERE=13
    AND=14
    INCLUDES=15
    EXCLUDES=16
    OR=17
    DYNAMIC=18
    OF=19
    SINCE=20
    OPERATOR=21
    NUMBER=22
    QUOTE=23
    WHITESPACE=24
    NEWLINE=25
    WORD=26

    def __init__(self, input, output=sys.stdout):
        super(MetricAlertConditionParser, self).__init__(input, output=output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class ExpressionContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.ExpressionContext, self).__init__(parent, invokingState)
            self.parser = parser

        def aggregation(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.AggregationContext,0)


        def operator(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.OperatorContext,0)


        def QUOTE(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.QUOTE)
            else:
                return self.getToken(MetricAlertConditionParser.QUOTE, i)

        def metric(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.MetricContext,0)


        def WHITESPACE(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WHITESPACE)
            else:
                return self.getToken(MetricAlertConditionParser.WHITESPACE, i)

        def threshold(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.ThresholdContext,0)


        def dynamics(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.DynamicsContext,0)


        def namespace(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.NamespaceContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.NamespaceContext,i)


        def dimensions(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.DimensionsContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.DimensionsContext,i)


        def NEWLINE(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.NEWLINE)
            else:
                return self.getToken(MetricAlertConditionParser.NEWLINE, i)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_expression

        def enterRule(self, listener):
            if hasattr(listener, "enterExpression"):
                listener.enterExpression(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitExpression"):
                listener.exitExpression(self)




    def expression(self):

        localctx = MetricAlertConditionParser.ExpressionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_expression)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 46
            self.aggregation()
            self.state = 52
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,0,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 47
                    self.namespace()
                    self.state = 48
                    self.match(MetricAlertConditionParser.T__0) 
                self.state = 54
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,0,self._ctx)

            self.state = 61
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [MetricAlertConditionParser.QUOTE]:
                self.state = 55
                self.match(MetricAlertConditionParser.QUOTE)
                self.state = 56
                self.metric()
                self.state = 57
                self.match(MetricAlertConditionParser.QUOTE)
                self.state = 58
                self.match(MetricAlertConditionParser.WHITESPACE)
                pass
            elif token in [MetricAlertConditionParser.T__0, MetricAlertConditionParser.T__1, MetricAlertConditionParser.T__2, MetricAlertConditionParser.T__3, MetricAlertConditionParser.T__4, MetricAlertConditionParser.T__5, MetricAlertConditionParser.T__6, MetricAlertConditionParser.T__7, MetricAlertConditionParser.T__8, MetricAlertConditionParser.NUMBER, MetricAlertConditionParser.WHITESPACE, MetricAlertConditionParser.WORD]:
                self.state = 60
                self.metric()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 63
            self.operator()
            self.state = 66
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [MetricAlertConditionParser.NUMBER]:
                self.state = 64
                self.threshold()
                pass
            elif token in [MetricAlertConditionParser.DYNAMIC]:
                self.state = 65
                self.dynamics()
                pass
            else:
                raise NoViableAltException(self)

            self.state = 72
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==MetricAlertConditionParser.WHITESPACE:
                self.state = 68
                self.match(MetricAlertConditionParser.WHITESPACE)
                self.state = 69
                self.dimensions()
                self.state = 74
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 78
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==MetricAlertConditionParser.NEWLINE:
                self.state = 75
                self.match(MetricAlertConditionParser.NEWLINE)
                self.state = 80
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AggregationContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.AggregationContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WORD(self):
            return self.getToken(MetricAlertConditionParser.WORD, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_aggregation

        def enterRule(self, listener):
            if hasattr(listener, "enterAggregation"):
                listener.enterAggregation(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitAggregation"):
                listener.exitAggregation(self)




    def aggregation(self):

        localctx = MetricAlertConditionParser.AggregationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_aggregation)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 81
            self.match(MetricAlertConditionParser.WORD)
            self.state = 82
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class NamespaceContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.NamespaceContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.NUMBER)
            else:
                return self.getToken(MetricAlertConditionParser.NUMBER, i)

        def WORD(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WORD)
            else:
                return self.getToken(MetricAlertConditionParser.WORD, i)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_namespace

        def enterRule(self, listener):
            if hasattr(listener, "enterNamespace"):
                listener.enterNamespace(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitNamespace"):
                listener.exitNamespace(self)




    def namespace(self):

        localctx = MetricAlertConditionParser.NamespaceContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_namespace)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 85 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 84
                    _la = self._input.LA(1)
                    if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MetricAlertConditionParser.T__0) | (1 << MetricAlertConditionParser.T__1) | (1 << MetricAlertConditionParser.NUMBER) | (1 << MetricAlertConditionParser.WORD))) != 0)):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()

                else:
                    raise NoViableAltException(self)
                self.state = 87 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,5,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class MetricContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.MetricContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.NUMBER)
            else:
                return self.getToken(MetricAlertConditionParser.NUMBER, i)

        def WORD(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WORD)
            else:
                return self.getToken(MetricAlertConditionParser.WORD, i)

        def WHITESPACE(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WHITESPACE)
            else:
                return self.getToken(MetricAlertConditionParser.WHITESPACE, i)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_metric

        def enterRule(self, listener):
            if hasattr(listener, "enterMetric"):
                listener.enterMetric(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitMetric"):
                listener.exitMetric(self)




    def metric(self):

        localctx = MetricAlertConditionParser.MetricContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_metric)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 90 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 89
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MetricAlertConditionParser.T__0) | (1 << MetricAlertConditionParser.T__1) | (1 << MetricAlertConditionParser.T__2) | (1 << MetricAlertConditionParser.T__3) | (1 << MetricAlertConditionParser.T__4) | (1 << MetricAlertConditionParser.T__5) | (1 << MetricAlertConditionParser.T__6) | (1 << MetricAlertConditionParser.T__7) | (1 << MetricAlertConditionParser.T__8) | (1 << MetricAlertConditionParser.NUMBER) | (1 << MetricAlertConditionParser.WHITESPACE) | (1 << MetricAlertConditionParser.WORD))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 92 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MetricAlertConditionParser.T__0) | (1 << MetricAlertConditionParser.T__1) | (1 << MetricAlertConditionParser.T__2) | (1 << MetricAlertConditionParser.T__3) | (1 << MetricAlertConditionParser.T__4) | (1 << MetricAlertConditionParser.T__5) | (1 << MetricAlertConditionParser.T__6) | (1 << MetricAlertConditionParser.T__7) | (1 << MetricAlertConditionParser.T__8) | (1 << MetricAlertConditionParser.NUMBER) | (1 << MetricAlertConditionParser.WHITESPACE) | (1 << MetricAlertConditionParser.WORD))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class OperatorContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.OperatorContext, self).__init__(parent, invokingState)
            self.parser = parser

        def OPERATOR(self):
            return self.getToken(MetricAlertConditionParser.OPERATOR, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_operator

        def enterRule(self, listener):
            if hasattr(listener, "enterOperator"):
                listener.enterOperator(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitOperator"):
                listener.exitOperator(self)




    def operator(self):

        localctx = MetricAlertConditionParser.OperatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_operator)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 94
            self.match(MetricAlertConditionParser.OPERATOR)
            self.state = 95
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ThresholdContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.ThresholdContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(MetricAlertConditionParser.NUMBER, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_threshold

        def enterRule(self, listener):
            if hasattr(listener, "enterThreshold"):
                listener.enterThreshold(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitThreshold"):
                listener.exitThreshold(self)




    def threshold(self):

        localctx = MetricAlertConditionParser.ThresholdContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_threshold)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 97
            self.match(MetricAlertConditionParser.NUMBER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DynamicContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.DynamicContext, self).__init__(parent, invokingState)
            self.parser = parser

        def DYNAMIC(self):
            return self.getToken(MetricAlertConditionParser.DYNAMIC, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dynamic

        def enterRule(self, listener):
            if hasattr(listener, "enterDynamic"):
                listener.enterDynamic(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDynamic"):
                listener.exitDynamic(self)




    def dynamic(self):

        localctx = MetricAlertConditionParser.DynamicContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_dynamic)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 99
            self.match(MetricAlertConditionParser.DYNAMIC)
            self.state = 100
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DynamicsContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.DynamicsContext, self).__init__(parent, invokingState)
            self.parser = parser

        def dynamic(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.DynamicContext,0)


        def dyn_sensitivity(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dyn_sensitivityContext,0)


        def dyn_violations(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dyn_violationsContext,0)


        def dyn_of_separator(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dyn_of_separatorContext,0)


        def dyn_windows(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dyn_windowsContext,0)


        def WHITESPACE(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WHITESPACE)
            else:
                return self.getToken(MetricAlertConditionParser.WHITESPACE, i)

        def dyn_since_seperator(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.Dyn_since_seperatorContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.Dyn_since_seperatorContext,i)


        def dyn_datetime(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.Dyn_datetimeContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.Dyn_datetimeContext,i)


        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dynamics

        def enterRule(self, listener):
            if hasattr(listener, "enterDynamics"):
                listener.enterDynamics(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDynamics"):
                listener.exitDynamics(self)




    def dynamics(self):

        localctx = MetricAlertConditionParser.DynamicsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_dynamics)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 102
            self.dynamic()
            self.state = 103
            self.dyn_sensitivity()
            self.state = 104
            self.dyn_violations()
            self.state = 105
            self.dyn_of_separator()
            self.state = 106
            self.dyn_windows()
            self.state = 113
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,7,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 107
                    self.match(MetricAlertConditionParser.WHITESPACE)
                    self.state = 108
                    self.dyn_since_seperator()
                    self.state = 109
                    self.dyn_datetime() 
                self.state = 115
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,7,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dyn_sensitivityContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dyn_sensitivityContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WORD(self):
            return self.getToken(MetricAlertConditionParser.WORD, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dyn_sensitivity

        def enterRule(self, listener):
            if hasattr(listener, "enterDyn_sensitivity"):
                listener.enterDyn_sensitivity(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDyn_sensitivity"):
                listener.exitDyn_sensitivity(self)




    def dyn_sensitivity(self):

        localctx = MetricAlertConditionParser.Dyn_sensitivityContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_dyn_sensitivity)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 116
            self.match(MetricAlertConditionParser.WORD)
            self.state = 117
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dyn_violationsContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dyn_violationsContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(MetricAlertConditionParser.NUMBER, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dyn_violations

        def enterRule(self, listener):
            if hasattr(listener, "enterDyn_violations"):
                listener.enterDyn_violations(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDyn_violations"):
                listener.exitDyn_violations(self)




    def dyn_violations(self):

        localctx = MetricAlertConditionParser.Dyn_violationsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_dyn_violations)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 119
            self.match(MetricAlertConditionParser.NUMBER)
            self.state = 120
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dyn_of_separatorContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dyn_of_separatorContext, self).__init__(parent, invokingState)
            self.parser = parser

        def OF(self):
            return self.getToken(MetricAlertConditionParser.OF, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dyn_of_separator

        def enterRule(self, listener):
            if hasattr(listener, "enterDyn_of_separator"):
                listener.enterDyn_of_separator(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDyn_of_separator"):
                listener.exitDyn_of_separator(self)




    def dyn_of_separator(self):

        localctx = MetricAlertConditionParser.Dyn_of_separatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_dyn_of_separator)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 122
            self.match(MetricAlertConditionParser.OF)
            self.state = 123
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dyn_windowsContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dyn_windowsContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(MetricAlertConditionParser.NUMBER, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dyn_windows

        def enterRule(self, listener):
            if hasattr(listener, "enterDyn_windows"):
                listener.enterDyn_windows(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDyn_windows"):
                listener.exitDyn_windows(self)




    def dyn_windows(self):

        localctx = MetricAlertConditionParser.Dyn_windowsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_dyn_windows)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 125
            self.match(MetricAlertConditionParser.NUMBER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dyn_since_seperatorContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dyn_since_seperatorContext, self).__init__(parent, invokingState)
            self.parser = parser

        def SINCE(self):
            return self.getToken(MetricAlertConditionParser.SINCE, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dyn_since_seperator

        def enterRule(self, listener):
            if hasattr(listener, "enterDyn_since_seperator"):
                listener.enterDyn_since_seperator(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDyn_since_seperator"):
                listener.exitDyn_since_seperator(self)




    def dyn_since_seperator(self):

        localctx = MetricAlertConditionParser.Dyn_since_seperatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_dyn_since_seperator)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 127
            self.match(MetricAlertConditionParser.SINCE)
            self.state = 128
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dyn_datetimeContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dyn_datetimeContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.NUMBER)
            else:
                return self.getToken(MetricAlertConditionParser.NUMBER, i)

        def WORD(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WORD)
            else:
                return self.getToken(MetricAlertConditionParser.WORD, i)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dyn_datetime

        def enterRule(self, listener):
            if hasattr(listener, "enterDyn_datetime"):
                listener.enterDyn_datetime(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDyn_datetime"):
                listener.exitDyn_datetime(self)




    def dyn_datetime(self):

        localctx = MetricAlertConditionParser.Dyn_datetimeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_dyn_datetime)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 131 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 130
                _la = self._input.LA(1)
                if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MetricAlertConditionParser.T__0) | (1 << MetricAlertConditionParser.T__4) | (1 << MetricAlertConditionParser.T__6) | (1 << MetricAlertConditionParser.T__9) | (1 << MetricAlertConditionParser.NUMBER) | (1 << MetricAlertConditionParser.WORD))) != 0)):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 133 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MetricAlertConditionParser.T__0) | (1 << MetricAlertConditionParser.T__4) | (1 << MetricAlertConditionParser.T__6) | (1 << MetricAlertConditionParser.T__9) | (1 << MetricAlertConditionParser.NUMBER) | (1 << MetricAlertConditionParser.WORD))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class WhereContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.WhereContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WHERE(self):
            return self.getToken(MetricAlertConditionParser.WHERE, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_where

        def enterRule(self, listener):
            if hasattr(listener, "enterWhere"):
                listener.enterWhere(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitWhere"):
                listener.exitWhere(self)




    def where(self):

        localctx = MetricAlertConditionParser.WhereContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_where)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 135
            self.match(MetricAlertConditionParser.WHERE)
            self.state = 136
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DimensionsContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.DimensionsContext, self).__init__(parent, invokingState)
            self.parser = parser

        def where(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.WhereContext,0)


        def dimension(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.DimensionContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.DimensionContext,i)


        def dim_separator(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.Dim_separatorContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.Dim_separatorContext,i)


        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dimensions

        def enterRule(self, listener):
            if hasattr(listener, "enterDimensions"):
                listener.enterDimensions(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDimensions"):
                listener.exitDimensions(self)




    def dimensions(self):

        localctx = MetricAlertConditionParser.DimensionsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_dimensions)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 138
            self.where()
            self.state = 139
            self.dimension()
            self.state = 145
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==MetricAlertConditionParser.T__7 or _la==MetricAlertConditionParser.AND:
                self.state = 140
                self.dim_separator()
                self.state = 141
                self.dimension()
                self.state = 147
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class DimensionContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.DimensionContext, self).__init__(parent, invokingState)
            self.parser = parser

        def dim_name(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dim_nameContext,0)


        def dim_operator(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dim_operatorContext,0)


        def dim_values(self):
            return self.getTypedRuleContext(MetricAlertConditionParser.Dim_valuesContext,0)


        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dimension

        def enterRule(self, listener):
            if hasattr(listener, "enterDimension"):
                listener.enterDimension(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDimension"):
                listener.exitDimension(self)




    def dimension(self):

        localctx = MetricAlertConditionParser.DimensionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_dimension)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 148
            self.dim_name()
            self.state = 149
            self.dim_operator()
            self.state = 150
            self.dim_values()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dim_separatorContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dim_separatorContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def AND(self):
            return self.getToken(MetricAlertConditionParser.AND, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dim_separator

        def enterRule(self, listener):
            if hasattr(listener, "enterDim_separator"):
                listener.enterDim_separator(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDim_separator"):
                listener.exitDim_separator(self)




    def dim_separator(self):

        localctx = MetricAlertConditionParser.Dim_separatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_dim_separator)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 152
            _la = self._input.LA(1)
            if not(_la==MetricAlertConditionParser.T__7 or _la==MetricAlertConditionParser.AND):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 153
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dim_operatorContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dim_operatorContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def INCLUDES(self):
            return self.getToken(MetricAlertConditionParser.INCLUDES, 0)

        def EXCLUDES(self):
            return self.getToken(MetricAlertConditionParser.EXCLUDES, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dim_operator

        def enterRule(self, listener):
            if hasattr(listener, "enterDim_operator"):
                listener.enterDim_operator(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDim_operator"):
                listener.exitDim_operator(self)




    def dim_operator(self):

        localctx = MetricAlertConditionParser.Dim_operatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 36, self.RULE_dim_operator)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 155
            _la = self._input.LA(1)
            if not(_la==MetricAlertConditionParser.INCLUDES or _la==MetricAlertConditionParser.EXCLUDES):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 156
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dim_val_separatorContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dim_val_separatorContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def OR(self):
            return self.getToken(MetricAlertConditionParser.OR, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dim_val_separator

        def enterRule(self, listener):
            if hasattr(listener, "enterDim_val_separator"):
                listener.enterDim_val_separator(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDim_val_separator"):
                listener.exitDim_val_separator(self)




    def dim_val_separator(self):

        localctx = MetricAlertConditionParser.Dim_val_separatorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 38, self.RULE_dim_val_separator)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 158
            _la = self._input.LA(1)
            if not(_la==MetricAlertConditionParser.T__7 or _la==MetricAlertConditionParser.OR):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 159
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dim_nameContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dim_nameContext, self).__init__(parent, invokingState)
            self.parser = parser

        def WORD(self):
            return self.getToken(MetricAlertConditionParser.WORD, 0)

        def WHITESPACE(self):
            return self.getToken(MetricAlertConditionParser.WHITESPACE, 0)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dim_name

        def enterRule(self, listener):
            if hasattr(listener, "enterDim_name"):
                listener.enterDim_name(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDim_name"):
                listener.exitDim_name(self)




    def dim_name(self):

        localctx = MetricAlertConditionParser.Dim_nameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 40, self.RULE_dim_name)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 161
            self.match(MetricAlertConditionParser.WORD)
            self.state = 162
            self.match(MetricAlertConditionParser.WHITESPACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dim_valuesContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dim_valuesContext, self).__init__(parent, invokingState)
            self.parser = parser

        def dim_value(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.Dim_valueContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.Dim_valueContext,i)


        def dim_val_separator(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(MetricAlertConditionParser.Dim_val_separatorContext)
            else:
                return self.getTypedRuleContext(MetricAlertConditionParser.Dim_val_separatorContext,i)


        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dim_values

        def enterRule(self, listener):
            if hasattr(listener, "enterDim_values"):
                listener.enterDim_values(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDim_values"):
                listener.exitDim_values(self)




    def dim_values(self):

        localctx = MetricAlertConditionParser.Dim_valuesContext(self, self._ctx, self.state)
        self.enterRule(localctx, 42, self.RULE_dim_values)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 164
            self.dim_value()
            self.state = 170
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,10,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 165
                    self.dim_val_separator()
                    self.state = 166
                    self.dim_value() 
                self.state = 172
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,10,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Dim_valueContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(MetricAlertConditionParser.Dim_valueContext, self).__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.NUMBER)
            else:
                return self.getToken(MetricAlertConditionParser.NUMBER, i)

        def WORD(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WORD)
            else:
                return self.getToken(MetricAlertConditionParser.WORD, i)

        def WHITESPACE(self, i=None):
            if i is None:
                return self.getTokens(MetricAlertConditionParser.WHITESPACE)
            else:
                return self.getToken(MetricAlertConditionParser.WHITESPACE, i)

        def getRuleIndex(self):
            return MetricAlertConditionParser.RULE_dim_value

        def enterRule(self, listener):
            if hasattr(listener, "enterDim_value"):
                listener.enterDim_value(self)

        def exitRule(self, listener):
            if hasattr(listener, "exitDim_value"):
                listener.exitDim_value(self)




    def dim_value(self):

        localctx = MetricAlertConditionParser.Dim_valueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 44, self.RULE_dim_value)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 174 
            self._errHandler.sync(self)
            _alt = 1
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 173
                    _la = self._input.LA(1)
                    if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << MetricAlertConditionParser.T__0) | (1 << MetricAlertConditionParser.T__2) | (1 << MetricAlertConditionParser.T__4) | (1 << MetricAlertConditionParser.T__5) | (1 << MetricAlertConditionParser.T__6) | (1 << MetricAlertConditionParser.T__7) | (1 << MetricAlertConditionParser.T__8) | (1 << MetricAlertConditionParser.T__10) | (1 << MetricAlertConditionParser.T__11) | (1 << MetricAlertConditionParser.NUMBER) | (1 << MetricAlertConditionParser.WHITESPACE) | (1 << MetricAlertConditionParser.WORD))) != 0)):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()

                else:
                    raise NoViableAltException(self)
                self.state = 176 
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,11,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





